from .height import get_screen_height
from .width import get_screen_width


def get_screen_size(driver: object) -> tuple[int, int]:
    width = get_screen_width(driver)
    height = get_screen_height(driver)
    return width, height



from .... import helper
from ....constant import timeout
from ....model.type.xpath import XPath
from ...check_if.is_clickable import check_if_is_clickable


def wait_until_element_is_clickable(driver: object, xpath: str, timeout: float = timeout.DEFAULT) -> None:
    xpath = XPath(xpath)
    helper.retry.until_condition_is_true(driver, xpath, func=check_if_is_clickable, timeout=timeout)



class URLSyntaxError(Exception):
    __slots__ = ["message"]

    def __init__(self, url: str) -> None:
        self.message = f"Invalid URL syntax: {url}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



from .....constant import timeout
from .....model.browser.base.driver import BrowserDriver
from .....model.browser.base.settings import BrowserSettings
from .....model.driver_methods import DriverMethods
from .changes import wait_until_url_changes
from .contains import wait_until_url_contains
from .equals import wait_until_url_equals


class WaitUntilUrlDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def changes(self, baseline_url: str, timeout: float = timeout.DEFAULT) -> None:
        """Wait until the browser URL has changed from a baseline URL, e.g. after a redirect or form action. The URL is evaluated as an exact match."""

        wait_until_url_changes(self._driver, baseline_url, timeout)

    def contains(self, url_fragment: str, timeout: float = timeout.DEFAULT) -> None:
        """Wait until the browser URL has changed, e.g. after a redirect. The URL variable can contain both a fragment (e.g. ?login=true) or a full URL (e.g. https://www.example.com/?login=true)."""

        wait_until_url_contains(self._driver, url_fragment, timeout)

    def equals(self, url: str, timeout: float = timeout.DEFAULT) -> None:
        """Wait until the browser URL has changed, e.g. after a redirect. The URL is evaluated as an exact match."""

        wait_until_url_equals(self._driver, url, timeout)



import re

from selenium.webdriver.common.by import By

from ...model.type.xpath import XPath


def check_if_contains_text(driver: object, xpath: str, regex: str, ignore_case: bool = True) -> bool:
    xpath = XPath(xpath)
    try:
        text = str(driver.find_element(By.XPATH, xpath).text)  # type: ignore
        if ignore_case:
            match = re.search(regex, text, re.IGNORECASE)
        else:
            match = re.search(regex, text)
        return bool(match)
    except Exception:
        return False



import asyncio

from ...browser.scroll.check_if.is_end_of_page import check_if_scroll_is_end_of_page
from ...browser.scroll.get.position import get_scroll_position
from ...browser.scroll.page.down import scroll_page_down
from ...browser.scroll.page.to_top import scroll_to_top_of_page
from ...browser.scroll.to_position import scroll_to_position
from ...model.screenshot import ScreenshotTempDataHandler


def firefox(driver: object, destination_file_path: str) -> None:
    driver.get_full_page_screenshot_as_file(destination_file_path)  # type: ignore


async def default(driver: object, destination_file_path: str, destination_dir: str, delay_seconds: float) -> None:
    async def async_scroll_page_down(driver: object, delay_seconds: float) -> None:
        scroll_page_down(driver, delay_seconds)

    async def async_scroll_to_position(driver: object, x: int, y: int, delay_seconds: float) -> None:
        scroll_to_position(driver, x_inital, y_initial, delay_seconds)

    async def get_screenshot_of_visible_portion_and_scroll_down(driver: object, handler: ScreenshotTempDataHandler, delay_seconds: float) -> None:
        task_save_screenshot = asyncio.create_task(handler.save_screenshot(driver))
        task_scroll_page_down = asyncio.create_task(async_scroll_page_down(driver, delay_seconds))
        await asyncio.gather(task_save_screenshot, task_scroll_page_down)
        handler.increment_iteration()

    # Save inital scroll position so we can return to it later.
    x_inital, y_initial = get_scroll_position(driver)

    # Prepare for iteration from the top of the page.
    scroll_to_top_of_page(driver, delay_seconds)
    handler = ScreenshotTempDataHandler(destination_dir=destination_dir, destination_file_path=destination_file_path)

    # Take screenshots of the visible portion until we reach the end of the page.
    await get_screenshot_of_visible_portion_and_scroll_down(driver, handler, delay_seconds)
    while check_if_scroll_is_end_of_page(driver) is not True:
        await get_screenshot_of_visible_portion_and_scroll_down(driver, handler, delay_seconds)

    # Merge screenshots, return to initial scroll position, and tidy up temp files.
    task_merge_temp_files_into_final_screenshot = asyncio.create_task(handler.merge_temp_files_into_final_screenshot())
    task_scroll_to_position = asyncio.create_task(async_scroll_to_position(driver, x_inital, y_initial, delay_seconds))
    await asyncio.gather(task_merge_temp_files_into_final_screenshot, task_scroll_to_position)
    handler.remove_temp_files()



def get_window_size(driver: object) -> tuple[int, int]:
    width = int(driver.get_window_size().get("width"))  # type: ignore
    height = int(driver.get_window_size().get("height"))  # type: ignore
    return width, height



from ... import constant, helper
from ...model.browser.base.settings import BrowserSettings
from ...model.screenshot import ScreenshotType


def mediate_file_name(file_name: str | None, screenshot_type: ScreenshotType | None = None) -> str:
    return helper.screenshot.file.get_default_name(screenshot_type) if file_name is None else file_name


def mediate_destination_dir(settings: BrowserSettings, destination_dir: str | None = None) -> str:
    if destination_dir is None:
        return settings.screenshot_dir
    helper.directory.create_if_not_exists(destination_dir)
    return helper.directory.ensure_trailing_slash(destination_dir)


def mediate_temp_dir(destination_dir: str) -> str:
    """As the temporary directory will be a sub directory to the destination, this assumes that the destination directory has been defined by the controller."""

    destination_dir = helper.directory.ensure_trailing_slash(destination_dir)
    temp_dir = f"{destination_dir}{constant.screenshot.TEMP_DIR}/"
    helper.directory.create_if_not_exists(temp_dir)
    return temp_dir



from ...model.window.controller import WindowHandleController
from .handle.current import get_current_window_handle


def window_close(driver: object, controller: WindowHandleController) -> None:
    current_window_handle_id = get_current_window_handle(driver)
    controller.remove_handle_by_id(current_window_handle_id)
    driver.close()  # type: ignore



from __future__ import annotations

from ...exception.file_png import FilePNGSyntaxError


class FilePNG(str):
    """Class to handle and validate PNG image files as "tiny type"."""

    __slots__ = ["value"]

    def __new__(cls, file_name: str | FilePNG) -> FilePNG:
        # If input already is a validated PNG image file, bypass and don't create a new object:
        return file_name if isinstance(file_name, FilePNG) else super().__new__(cls, file_name)

    def __init__(self, file_name: str) -> None:
        self.value: str = file_name
        if not self.is_valid():
            raise FilePNGSyntaxError(file_name)

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value

    def is_valid(self) -> bool:
        return bool(self.value.endswith(".png"))



from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By

from ...constant import timeout
from ...exception.element import NoElementDimensionsFoundException
from ...exception.timeout import WaitForElementTimeoutException
from ...model.type.xpath import XPath
from ..wait.for_element import wait_for_element


def get_dimensions(driver: object, xpath: str, timeout: float = timeout.DEFAULT) -> tuple[int, int]:
    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    try:
        # Returns dictionary object, e.g. {'height': 598, 'width': 479}:
        dimensions: dict[str, int] = driver.find_element(By.XPATH, xpath).size  # type: ignore
        return dimensions.get("width") or 0, dimensions.get("height") or 0
    except TimeoutException:
        raise WaitForElementTimeoutException(driver, xpath) from TimeoutException
    except NoSuchElementException:
        raise NoElementDimensionsFoundException(driver, xpath) from NoSuchElementException



import random
import time

from ...constant import timeout


def wait_random_time(min_seconds: float = timeout.VERY_SHORT, max_seconds: float = timeout.DEFAULT) -> None:
    time.sleep(random.uniform(min_seconds, max_seconds))



import time


def scroll_by(driver: object, x: int, y: int, delay_seconds: float) -> None:
    driver.execute_script(f"window.scrollBy({x}, {y});")  # type: ignore
    time.sleep(delay_seconds)  # Small delay to ensure that the screen is updated after scroll.



from selenium import webdriver

from ... import factory
from .base.driver import BrowserDriver
from .base.type import BrowserType


class FirefoxBrowserDriver(BrowserDriver):
    def ensure_browser_type(self) -> None:
        self.settings.type = BrowserType.FIREFOX

    def set_webdriver(self) -> object:
        if self.settings.path_to_executable is None:
            return webdriver.Firefox(  # type: ignore
                options=self.firefox_options)
        else:
            return webdriver.Firefox(  # type: ignore
                executable_path=self.settings.path_to_executable,
                options=self.firefox_options)

    def disable_images(self) -> None:
        if self.settings.disable_images:
            self.firefox_options.set_preference("permissions.default.image", 2)
            self.firefox_options.set_preference("dom.ipc.plugins.enabled.libflashplayer.so", "false")

    def enable_headless(self) -> None:
        if self.settings.headless:
            self.firefox_options.add_argument("--headless")  # type: ignore

    def set_page_load_strategy(self) -> None:
        self.firefox_options = factory.set.page_load_strategy(self, self.firefox_options)  # type: ignore



from selenium.webdriver.common.by import By

from ....constant import timeout
from ....model.type.xpath import XPath
from ...wait.for_element import wait_for_element


def get_attribute_values(driver: object, xpath: str, attribute: str, timeout: float = timeout.DEFAULT) -> list[str]:
    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    elements: list[object] = driver.find_elements(By.XPATH, xpath)  # type: ignore
    return [element.get_attribute(attribute) for element in elements]  # type: ignore



from dataclasses import dataclass


@dataclass(kw_only=True, slots=True)
class LoginCredentials:
    """Object to define a user profile's access credentials.

    Separated from the login form so that the multiple users/roles can use the same form."""

    username: str
    password: str



from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait  # type: ignore

from ... import constant
from ...constant import timeout
from ...exception.element import NoElementFoundException
from ...exception.timeout import WaitForElementTimeoutException
from ...model.type.xpath import XPath


def wait_for_element(driver: object, xpath: str, timeout: float = timeout.DEFAULT) -> None:
    xpath = XPath(xpath)
    if timeout == constant.timeout.BYPASS:
        return
    try:
        WebDriverWait(driver, timeout).until(EC.presence_of_element_located((By.XPATH, xpath)))  # type: ignore
    except TimeoutException:
        raise WaitForElementTimeoutException(driver, xpath) from TimeoutException
    except NoSuchElementException:
        raise NoElementFoundException(driver, xpath) from NoSuchElementException



import time

from ..constant import interval, timeout
from ..exception.retry import RetryTimeoutException
from ..model.type.callable import DriverGetBoolCallable, DriverGetTextCallable


def calculate_number_of_retries(total_time: int | float, interval: int | float) -> int:
    return int(total_time // interval)


def get_text(driver: object, input: str, func: DriverGetTextCallable, timeout: float = timeout.DEFAULT, wait_interval_seconds: float = interval.DEFAULT) -> str:
    text = func(driver, input)
    retries_left = calculate_number_of_retries(timeout, wait_interval_seconds)
    while not text and retries_left > 0:
        time.sleep(wait_interval_seconds)
        text = func(driver, input)
        retries_left -= 1
        if retries_left == 0:
            raise RetryTimeoutException(func)
    return text


def until_condition_is_true(driver: object, *args: str | list[object], func: DriverGetBoolCallable, timeout: float = timeout.DEFAULT, wait_interval_seconds: float = interval.DEFAULT) -> None:
    retries_left = calculate_number_of_retries(timeout, wait_interval_seconds)
    while func(driver, *args) is False and retries_left > 0:
        time.sleep(wait_interval_seconds)
        retries_left -= 1
        if retries_left == 0:
            raise RetryTimeoutException(func)


def until_condition_is_false(driver: object, *args: str | list[object], func: DriverGetBoolCallable, timeout: float = timeout.DEFAULT, wait_interval_seconds: float = interval.DEFAULT) -> None:
    retries_left = calculate_number_of_retries(timeout, wait_interval_seconds)
    while func(driver, *args) is True and retries_left > 0:
        time.sleep(wait_interval_seconds)
        retries_left -= 1
        if retries_left == 0:
            raise RetryTimeoutException(func)



def get_window_position(driver: object) -> tuple[int, int]:
    x = int(driver.get_window_position().get("x"))  # type: ignore
    y = int(driver.get_window_position().get("y"))  # type: ignore
    return x, y



import time

from ...model.combo_settings.login_credentials import LoginCredentials
from ...model.combo_settings.login_form import LoginForm1Step, LoginForm2Steps
from ..click.button import click_button
from ..input.value import input_value
from ..open.url_if_not_current import open_url_if_not_current
from ..wait.for_element import wait_for_element
from ..wait.until.url.contains import wait_until_url_contains


def combo_log_in(driver: object, login_credentials: LoginCredentials, login_form: LoginForm1Step | LoginForm2Steps) -> None:
    def login_form_1_step(login_form: LoginForm1Step) -> None:
        input_value(driver, login_form.username_input_xpath, login_credentials.username)
        input_value(driver, login_form.password_input_xpath, login_credentials.password)
        click_button(driver, login_form.submit_button_xpath)

    def login_form_2_steps(login_form: LoginForm2Steps) -> None:
        input_value(driver, login_form.username_input_xpath, login_credentials.username)
        click_button(driver, login_form.username_submit_button_xpath)
        input_value(driver, login_form.password_input_xpath, login_credentials.password)
        click_button(driver, login_form.password_submit_button_xpath)

    if login_form.url is not None:
        open_url_if_not_current(driver, login_form.url)

    if type(login_form) is LoginForm1Step:
        login_form_1_step(login_form)
    elif type(login_form) is LoginForm2Steps:
        login_form_2_steps(login_form)

    if login_form.post_login_wait_seconds is not None:
        time.sleep(login_form.post_login_wait_seconds)
    if login_form.post_login_url_contains is not None:
        wait_until_url_contains(driver, login_form.post_login_url_contains)
    if login_form.post_login_element_xpath is not None:
        wait_for_element(driver, login_form.post_login_element_xpath)



__all__ = ["complete_page", "controller", "file", "save", "save_element"]


from ... import helper
from . import complete_page, controller, file


def save(driver: object, file_path: str) -> None:
    """Take screenshot of visible portion. Reference: https://www.selenium.dev/documentation/webdriver/browser/windows/#takescreenshot"""

    driver.save_screenshot(file_path)  # type: ignore


def save_element(element: object, file_path: str) -> None:
    """Take screenshot of element. Reference: https://www.selenium.dev/documentation/webdriver/browser/windows/#takeelementscreenshot"""

    element.screenshot(file_path)  # type: ignore


def merge_images(all_temp_file_paths: list[str], save_file_path: str) -> None:
    if not all_temp_file_paths:
        return
    merged_image = helper.image.open(all_temp_file_paths[0])
    if len(all_temp_file_paths) > 1:
        for file_path in all_temp_file_paths[1:]:
            image_add = helper.image.open(file_path)
            merged_image = helper.image.merge_vertically(merged_image, image_add)
    helper.image.save(merged_image, save_file_path)



from ..... import helper
from .....constant import timeout
from .....model.type.xpath import XPath
from ....get.text import get_text
from ...for_element import wait_for_element


def wait_until_text_changes(driver: object, xpath: str, baseline_text: str, timeout: float = timeout.DEFAULT) -> None:
    def has_text_changed(driver: object, baseline_text: str) -> bool:
        return get_text(driver, xpath) != baseline_text

    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    helper.retry.until_condition_is_true(driver, baseline_text, func=has_text_changed, timeout=timeout)



from PIL import Image  # type: ignore


def is_element_loaded(driver: object, element: object) -> bool:
    """Check if image element is loaded and ready in the DOM."""

    is_image_loaded: bool = driver.execute_script(  # type: ignore
        "return arguments[0].complete && typeof arguments[0].naturalWidth != 'undefined' && arguments[0].naturalWidth > 0;",
        element)
    return is_image_loaded or False


def merge_vertically(image_base: Image, image_add: Image) -> Image:  # type: ignore
    """Merge two images vertically. Assumes both images have the same width. "image_add" will be added below "image_base"."""

    merged_image_width: int = image_base.width
    merged_image_height: int = image_base.height + image_add.height
    merged_image = Image.new("RGB", (merged_image_width, merged_image_height))
    merged_image.paste(image_base, (0, 0))
    merged_image.paste(image_add, (0, image_base.height))
    return merged_image


def open(file_path: str) -> Image:  # type: ignore
    """Open image from file path."""

    return Image.open(file_path)


def save(image: Image, file_path: str) -> Image:  # type: ignore
    """Save image to file path."""

    return image.save(file_path)



from ...constant import timeout
from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .hover import mouse_hover


class MouseDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def hover(self, xpath: str, timeout: float = timeout.DEFAULT) -> None:
        """Simulate moving the mouse cursor over the middle of an element."""

        mouse_hover(self._driver, xpath, timeout)



from selenium.webdriver.common.by import By

from ...model.type.xpath import XPath


def check_if_contains_any_text(driver: object, xpath: str) -> bool:
    xpath = XPath(xpath)
    try:
        text = str(driver.find_element(By.XPATH, xpath).text)  # type: ignore
        return bool(text)
    except Exception:
        return False



from selenium.webdriver.common.by import By

from ....constant import timeout
from ....model.type.xpath import XPath
from ...wait.for_element import wait_for_element


def get_attribute_value(driver: object, xpath: str, attribute: str, timeout: float = timeout.DEFAULT) -> str:
    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    return driver.find_element(By.XPATH, xpath).get_attribute(attribute)  # type: ignore



from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By

from ...constant import timeout
from ...exception.element import NoElementFoundException
from ...exception.timeout import WaitForElementTimeoutException
from ...model.type.xpath import XPath
from ..wait.for_element import wait_for_element


def click_button(driver: object, xpath: str, timeout: float = timeout.DEFAULT) -> None:
    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    try:
        element = driver.find_element(By.XPATH, xpath)  # type: ignore
        element.click()
    except TimeoutException:
        raise WaitForElementTimeoutException(driver, xpath) from TimeoutException
    except NoSuchElementException:
        raise NoElementFoundException(driver, xpath) from NoSuchElementException



from ... import helper
from ...model.browser.base.settings import BrowserSettings
from ...model.screenshot import ScreenshotType
from ...model.type.file_png import FilePNG


def get_screenshot_of_visible_portion(driver: object, settings: BrowserSettings, file_name: str | None = None, destination_dir: str | None = None) -> None:
    if file_name is not None:
        file_name = FilePNG(file_name)
    file_name = helper.screenshot.controller.mediate_file_name(file_name, ScreenshotType.VISIBLE_PORTION)
    destination_dir = helper.screenshot.controller.mediate_destination_dir(settings, destination_dir)
    file_path = helper.screenshot.file.get_path(destination_dir, file_name)
    helper.screenshot.save(driver, file_path)



from ... import constant, helper
from ...model.screenshot import ScreenshotType
from ..date_time import get_current_date_and_time, get_timestamp


def get_default_name(screenshot_type: ScreenshotType | None = None) -> str:
    """Example: \"Browserist screenshot 2022-02-12 at 22.12.34.png\""""

    date, time = get_current_date_and_time()
    appendix = "" if screenshot_type is None else f" ({screenshot_type.value})"
    return f"Browserist screenshot {date} at {time}{appendix}.png"


def get_temp_prefix_without_iterator_and_file_type() -> str:
    """Intended use: \"2022-02-12_22.12.34_temp_1.png\". Only returns the \"2022-02-12_22.12.34_temp\" part so remember to add the \"_1.png\" part."""

    timestamp = get_timestamp()
    return f"{timestamp}_{constant.screenshot.TEMP_FILE}"


def get_path(destination_dir: str, file_name: str) -> str:
    """Merge destination directory and file name into a single path. Assumes that the directory is valid and exists."""

    destination_dir = helper.directory.ensure_trailing_slash(destination_dir)
    return f"{destination_dir}{file_name}"



from ...constant import timeout
from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .by import scroll_by
from .check_if.__main__ import ScrollCheckIfDriverMethods
from .down_by import scroll_down_by
from .get.__main__ import ScrollGetDriverMethods
from .into_view import scroll_into_view
from .into_view_if_not_visible import scroll_into_view_if_not_visible
from .page.__main__ import ScrollPageDriverMethods
from .to_position import scroll_to_position
from .up_by import scroll_up_by


class ScrollDriverMethods(DriverMethods):
    __slots__ = ["check_if", "get", "page"]

    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)
        self.check_if: ScrollCheckIfDriverMethods = ScrollCheckIfDriverMethods(browser_driver, settings)
        self.get: ScrollGetDriverMethods = ScrollGetDriverMethods(browser_driver, settings)
        self.page: ScrollPageDriverMethods = ScrollPageDriverMethods(browser_driver, settings)

    def by(self, x: int, y: int, delay_seconds: float = 1) -> None:
        """If possible, scroll by X and Y pixels as relative position. Add custom delay in seconds to ensure the view is updated after scroll."""

        scroll_by(self._driver, x, y, delay_seconds)

    def down_by(self, pixels: int, delay_seconds: float = 1) -> None:
        """If possible, scroll down in pixels. Horisontal position is unchanged. Add custom delay in seconds to ensure the view is updated after scroll."""

        scroll_down_by(self._driver, pixels, delay_seconds)

    def into_view(self, xpath: str, timeout: float = timeout.DEFAULT, delay_seconds: float = 1) -> None:
        """Find element and scroll up or down until element is visible. Add custom delay in seconds to ensure the view is updated after scroll."""

        scroll_into_view(self._driver, xpath, timeout, delay_seconds)

    def into_view_if_not_visible(self, xpath: str, timeout: float = timeout.DEFAULT, delay_seconds: float = 1) -> None:
        """If not visible, find element and scroll up or down until element is visible. Add custom delay in seconds to ensure the view is updated after scroll."""

        scroll_into_view_if_not_visible(self._driver, xpath, timeout, delay_seconds)

    def to_position(self, x: int, y: int, delay_seconds: float = 1) -> None:
        """If possible, scroll to coordinate X and Y pixels of page as absolute position. Add custom delay in seconds to ensure the view is updated after scroll."""

        scroll_to_position(self._driver, x, y, delay_seconds)

    def up_by(self, pixels: int, delay_seconds: float = 1) -> None:
        """If possible, scroll up in pixels. Horisontal position is unchanged. Add custom delay in seconds to ensure the view is updated after scroll."""

        scroll_up_by(self._driver, pixels, delay_seconds)



def get_scroll_position(driver: object) -> tuple[int, int]:
    x = int(driver.execute_script("return window.pageXOffset;"))  # type: ignore
    y = int(driver.execute_script("return window.pageYOffset;"))  # type: ignore
    return x, y



import time

from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By

from ...exception.element import NoElementFoundException
from ...exception.timeout import WaitForElementTimeoutException
from ...model.type.xpath import XPath
from ..wait.for_element import wait_for_element


def scroll_into_view(driver: object, xpath: str, timeout: float, delay_seconds: float) -> None:
    xpath = XPath(xpath)
    try:
        wait_for_element(driver, xpath, timeout)
        element: object = driver.find_element(By.XPATH, xpath)  # type: ignore
        driver.execute_script("arguments[0].scrollIntoView();", element)  # type: ignore
        time.sleep(delay_seconds)  # Small delay to ensure that the screen is updated after scroll.
    except TimeoutException:
        raise WaitForElementTimeoutException(driver, xpath) from TimeoutException
    except NoSuchElementException:
        raise NoElementFoundException(driver, xpath) from NoSuchElementException



import asyncio

from ... import helper
from ...model.browser.base.settings import BrowserSettings
from ...model.browser.base.type import BrowserType
from ...model.screenshot import ScreenshotType
from ...model.type.file_png import FilePNG


def get_screenshot_of_complete_page(driver: object, settings: BrowserSettings, file_name: str | None = None, destination_dir: str | None = None, delay_seconds: float = 1) -> None:
    destination_dir = helper.screenshot.controller.mediate_destination_dir(settings, destination_dir)
    if file_name is not None:
        file_name = FilePNG(file_name)
    file_name = helper.screenshot.controller.mediate_file_name(file_name, ScreenshotType.COMPLETE_PAGE)
    destination_file_path = helper.screenshot.file.get_path(destination_dir, file_name)

    match settings.type:
        case BrowserType.FIREFOX:
            helper.screenshot.complete_page.firefox(driver, destination_file_path)
        case _:
            asyncio.run(
                helper.screenshot.complete_page.default(driver, destination_file_path, destination_dir, delay_seconds)
            )



from datetime import datetime


def get_timestamp() -> str:
    return datetime.now().strftime("%Y-%m-%d_%H.%M.%S")


def get_current_date_and_time() -> tuple[str, str]:
    """Returns current date and time as tuple. Usage:

    date, time = helper.date_time.get_current_date_and_time()"""

    timestamp = get_timestamp()
    timestamp_split = timestamp.split("_")
    return timestamp_split[0], timestamp_split[1]



from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from .is_end_of_page import check_if_scroll_is_end_of_page
from .is_top_of_page import check_if_scroll_is_top_of_page


class ScrollCheckIfDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def is_end_of_page(self) -> bool:
        """Check if current scroll position is at the end of the page."""

        return check_if_scroll_is_end_of_page(self._driver)

    def is_top_of_page(self) -> bool:
        """Check if current scroll position is at the top of the page."""

        return check_if_scroll_is_top_of_page(self._driver)



from .... import helper
from ....constant import timeout
from ....model.type.xpath import XPath
from ...check_if.is_displayed import check_if_is_displayed


def wait_until_element_disappears(driver: object, xpath: str, timeout: float = timeout.DEFAULT) -> None:
    xpath = XPath(xpath)
    helper.retry.until_condition_is_false(driver, xpath, func=check_if_is_displayed, timeout=timeout)



__all__ = ["date_time", "directory", "file", "image", "operating_system",
           "retry", "screenshot", "url", "window_handle", "xpath"]

from . import date_time, directory, file, image, operating_system, retry, screenshot, url, window_handle, xpath



from ..model.browser.base.driver import BrowserDriver
from ..model.browser.base.settings import BrowserSettings
from ..model.browser.base.type import BrowserType
from ..model.browser.chrome import ChromeBrowserDriver
from ..model.browser.edge import EdgeBrowserDriver
from ..model.browser.firefox import FirefoxBrowserDriver
from ..model.browser.internet_explorer import InternetExplorerBrowserDriver
from ..model.browser.opera import OperaBrowserDriver
from ..model.browser.safari import SafariBrowserDriver


def browser_driver(settings: BrowserSettings) -> BrowserDriver:
    match(settings.type):
        case BrowserType.CHROME:
            return ChromeBrowserDriver(settings)
        case BrowserType.EDGE:
            return EdgeBrowserDriver(settings)
        case BrowserType.FIREFOX:
            return FirefoxBrowserDriver(settings)
        case BrowserType.INTERNET_EXPLORER:
            return InternetExplorerBrowserDriver(settings)
        case BrowserType.OPERA:
            return OperaBrowserDriver(settings)
        case BrowserType.SAFARI:
            return SafariBrowserDriver(settings)
        case _:
            raise ValueError(settings.type)



from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .contains_any_text import check_if_contains_any_text
from .contains_text import check_if_contains_text
from .does_exist import check_if_does_exist
from .is_clickable import check_if_is_clickable
from .is_disabled import check_if_is_disabled
from .is_displayed import check_if_is_displayed
from .is_enabled import check_if_is_enabled
from .is_image_loaded import check_if_is_image_loaded


class CheckIfDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def contains_any_text(self, xpath: str) -> bool:
        """Check if element contains any text."""

        return check_if_contains_any_text(self._driver, xpath)

    def contains_text(self, xpath: str, regex: str, ignore_case: bool = True) -> bool:
        """Check if element contains text. The condition works for both ordinary text (e.g. "Submit") or regular expression (e.g. r"colou?r"). Note it's a search for text, not a strict text match."""

        return check_if_contains_text(self._driver, xpath, regex, ignore_case)

    def does_exist(self, xpath: str) -> bool:
        """Check if element exists."""

        return check_if_does_exist(self._driver, xpath)

    def is_clickable(self, xpath: str) -> bool:
        """Check if element is clickable."""

        return check_if_is_clickable(self._driver, xpath)

    def is_disabled(self, xpath: str) -> bool:
        """Check whether element is disabled."""

        return check_if_is_disabled(self._driver, xpath)

    def is_displayed(self, xpath: str) -> bool:
        """Check visibility status of an element."""

        return check_if_is_displayed(self._driver, xpath)

    def is_enabled(self, xpath: str) -> bool:
        """Check whether element is enabled."""

        return check_if_is_enabled(self._driver, xpath)

    def is_image_loaded(self, xpath: str) -> bool:
        """Check is image is loaded and ready in the DOM."""

        return check_if_is_image_loaded(self._driver, xpath)



from abc import ABC, abstractmethod

from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.ie.options import Options as IEOptions
from selenium.webdriver.safari.options import Options as SafariOptions

from .... import helper
from .settings import BrowserSettings
from .type import BrowserType


class BrowserDriver(ABC):
    """Abstract class that contains the Selenium web driver based on browser type and configuration."""

    __slots__ = ["settings", "chrome_options", "edge_options", "firefox_options",
                 "ie_options", "safari_options", "webdriver"]

    def __init__(self, settings: BrowserSettings) -> None:
        """Initiates basic properties of the Selenium web driver."""

        self.settings = settings
        helper.directory.create_if_not_exists(self.settings.screenshot_dir)

        match(self.settings.type):
            case BrowserType.CHROME | BrowserType.OPERA:
                self.chrome_options: ChromeOptions = ChromeOptions()
            case BrowserType.EDGE:
                self.edge_options: EdgeOptions = EdgeOptions()  # type: ignore
            case BrowserType.FIREFOX:
                self.firefox_options: FirefoxOptions = FirefoxOptions()  # type: ignore
            case BrowserType.INTERNET_EXPLORER:
                self.ie_options: IEOptions = IEOptions()  # type: ignore
            case BrowserType.SAFARI:
                self.safari_options: SafariOptions = SafariOptions()  # type: ignore

        self.ensure_browser_type()
        self.set_options_and_profile()
        self.webdriver: object = self.set_webdriver()

    @abstractmethod
    def ensure_browser_type(self) -> None:
        """Method to ensure the correct browser type if a specific browser instance is created directly from a subclass (e.g. FirefoxBrowserDriver) without the optional settings as argument, simply as Chrome is default browser."""

        raise NotImplementedError

    @abstractmethod
    def set_webdriver(self) -> object:
        """Method to set web driver based on the settings."""

        raise NotImplementedError

    def set_options_and_profile(self) -> None:
        """Internal task initialiser that runs the configuration methods to disable images, enable headless, etc."""

        self.disable_images()
        self.enable_headless()
        self.set_page_load_strategy()

    @abstractmethod
    def disable_images(self) -> None:
        """Method to configure web driver to disable download of images for faster browsing."""

        raise NotImplementedError

    @abstractmethod
    def enable_headless(self) -> None:
        """Method to enable headless version of web driver (i.e. don't open browser window) for faster browsing."""

        raise NotImplementedError

    @abstractmethod
    def set_page_load_strategy(self) -> None:
        """Method to set the page load strategy to define whether the web driver should wait until all assets are downloaded (slower) or not (faster)."""

        raise NotImplementedError



from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .complete_page import get_screenshot_of_complete_page
from .element import get_screenshot_of_element
from .visible_portion import get_screenshot_of_visible_portion


class ScreenshotDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def complete_page(self, file_name: str | None = None, destination_dir: str | None = None, delay_seconds: float = 1) -> None:
        """Take screenshot of complete page and save as PNG image. Default destination directory is from where the script is executed. Examples:

        browser.screenshot.complete_page() # Default file name and destination

        browser.screenshot.complete_page("image.png") # Custom file name and default destination

        browser.screenshot.complete_page("image.png", "./screenshots") # Custom file name and destination

        browser.screenshot.complete_page(destination_dir = "./screenshots") # Default file name and custom destination

        Firefox is recommended browser for complete page screenshots as it executes this in one go. Other browsers can't capture the entire page at once, and so we need to merge screenshots portion by portions – this is obviously much slower.

        Use "delay_seconds" to adjust iteration delay to ensure that the screen is updated after each scroll."""

        get_screenshot_of_complete_page(self._driver, self._settings, file_name, destination_dir, delay_seconds)

    def element(self, xpath: str, file_name: str | None = None, destination_dir: str | None = None) -> None:
        """Take screenshot of visible portion and save as PNG image. Default destination directory is from where the script is executed. Examples:

        browser.screenshot.element("/element/xpath") # Default file name and destination

        browser.screenshot.element("/element/xpath", "image.png") # Custom file name and default destination

        browser.screenshot.element("/element/xpath", "image.png", "./screenshots") # Custom file name and destination

        browser.screenshot.element("/element/xpath", destination_dir = "./screenshots") # Default file name and custom destination

        Note that "/element/xpath" should be a valid XPath expression."""

        get_screenshot_of_element(self._driver, xpath, self._settings, file_name, destination_dir)

    def visible_portion(self, file_name: str | None = None, destination_dir: str | None = None) -> None:
        """Take screenshot of visible portion and save as PNG image. Default destination directory is from where the script is executed. Examples:

        browser.screenshot.visible_portion() # Default file name and destination

        browser.screenshot.visible_portion("image.png") # Custom file name and default destination

        browser.screenshot.visible_portion("image.png", "./screenshots") # Custom file name and destination

        browser.screenshot.visible_portion(destination_dir = "./screenshots") # Default file name and custom destination"""

        get_screenshot_of_visible_portion(self._driver, self._settings, file_name, destination_dir)



from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait  # type: ignore

from ...constant import timeout
from ...model.type.xpath import XPath


def check_if_is_clickable(driver: object, xpath: str) -> bool:
    xpath = XPath(xpath)
    try:
        element: object = WebDriverWait(driver, timeout.BYPASS).until(
            EC.element_to_be_clickable((By.XPATH, xpath)))  # type: ignore
        return element is not None
    except (TimeoutException, NoSuchElementException, Exception):
        return False



from ..... import helper
from .....constant import timeout
from ....get.page_title import get_page_title


def wait_until_page_title_changes(driver: object, baseline_text: str, timeout: float = timeout.DEFAULT) -> None:
    def has_page_title_changed(driver: object, baseline_text: str) -> bool:
        return get_page_title(driver) != baseline_text

    helper.retry.until_condition_is_true(driver, baseline_text, func=has_page_title_changed, timeout=timeout)



from selenium.webdriver.common.by import By

from ...constant import timeout
from ..wait.for_element import wait_for_element


def get_elements_by_tag(driver: object, tag: str, timeout: float = timeout.DEFAULT) -> list[object]:
    wait_for_element(driver, f"//{tag}", timeout)
    return driver.find_elements(By.TAG_NAME, tag)  # type: ignore



from selenium.webdriver.common.by import By

from ... import helper
from ...constant import timeout
from ...model.type.xpath import XPath
from ..wait.for_element import wait_for_element


def get_text(driver: object, xpath: str, timeout: float = timeout.DEFAULT) -> str:
    def get_inner_text_of_element(driver: object, xpath: str) -> str:
        return str(driver.find_element(By.XPATH, xpath).text)  # type: ignore

    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    return helper.retry.get_text(driver, xpath, get_inner_text_of_element)



from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.combo_settings.cookie_banner import CookieBannerSettings
from ...model.combo_settings.login_credentials import LoginCredentials
from ...model.combo_settings.login_form import LoginForm1Step, LoginForm2Steps
from ...model.combo_settings.search import SearchSettings
from ...model.driver_methods import DriverMethods
from .cookie_banner import combo_cookie_banner
from .log_in import combo_log_in
from .search import combo_search


class ComboDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def cookie_banner(self, settings: CookieBannerSettings) -> None:
        """Standardised combination of methods to accept or decline cookies."""

        combo_cookie_banner(self._driver, settings)

    def log_in(self, login_credentials: LoginCredentials, login_form: LoginForm1Step | LoginForm2Steps) -> None:
        """Standardised combination of methods to log in.

        wait_seconds: Extra seconds in addition to timeout to make sure the login is processed and that the user is redirected succesfully."""

        combo_log_in(self._driver, login_credentials, login_form)

    def search(self, term: str, settings: SearchSettings) -> None:
        """Standardised combination of methods to perform search."""

        combo_search(self._driver, term, settings)



from ... import helper
from ...browser.window.handle.current import get_current_window_handle
from ...exception.window_handle import (WindowHandleIdNotFoundError, WindowHandleIdNotUniqueError,
                                        WindowHandleIdNotValidError, WindowHandleNameNotFoundError,
                                        WindowHandleNameNotUniqueError, WindowHandleNameNotValidError)
from .handle import WindowHandle


class WindowHandleController:
    __slots__ = ["_window_handles", "_original_window_name", "_counter"]

    def __init__(self, driver: object) -> None:
        self._counter: int = 1
        self._original_window_name = str(self._counter)
        self._window_handles: list[WindowHandle] = [
            WindowHandle(
                name=self._original_window_name,
                id=get_current_window_handle(driver),
            )]

    def add_handle(self, id: str, name: str | None = None) -> None:
        """Add new window handle to list. The name is optional."""

        if not helper.window_handle.is_valid_id(id):
            raise WindowHandleIdNotValidError(id)
        if helper.window_handle.id_already_exists(id, self._window_handles):
            raise WindowHandleIdNotUniqueError(id)
        self._counter += 1
        if name is None:
            name = str(self._counter)
        if helper.window_handle.name_already_exists(name, self._window_handles):
            raise WindowHandleNameNotUniqueError(name)
        self._window_handles.append(WindowHandle(name, id))

    def remove_handle_by_id(self, id: str) -> None:
        """Remove window handle ID."""

        if not helper.window_handle.is_valid_id(id):
            raise WindowHandleIdNotValidError(id)
        checksum = self.count()
        for window_handle in self._window_handles:
            if id == window_handle.id:
                self._window_handles.remove(window_handle)
        if self.count() != checksum - 1:
            raise WindowHandleIdNotFoundError(id)

    def remove_handle_by_name(self, name: str) -> None:
        """Remove window handle name."""

        if name == self._original_window_name:
            raise WindowHandleNameNotValidError(name)
        checksum = self.count()
        for window_handle in self._window_handles:
            if name == window_handle.name:
                self._window_handles.remove(window_handle)
        if self.count() != checksum - 1:
            raise WindowHandleNameNotFoundError(name)

    def get_handle_id_by_name(self, name: str) -> str:
        """Get window handle ID by name."""

        for window_handle in self._window_handles:
            if name == window_handle.name:
                return window_handle.id
        raise WindowHandleNameNotFoundError(name)

    def get_handle_name_by_id(self, id: str) -> str:
        """Get window handle name by ID."""

        for window_handle in self._window_handles:
            if id == window_handle.id:
                return window_handle.name
        raise WindowHandleIdNotFoundError(id)

    def count(self) -> int:
        """Get current total number of window handles."""

        return len(self._window_handles)



from dataclasses import dataclass
from enum import Enum, unique

from .. import helper
from .type.file_png import FilePNG


@unique
class ScreenshotType(Enum):
    """Class to define screenshot type, e.g. complete page, visible portion, element."""

    COMPLETE_PAGE = "complete page"
    ELEMENT = "element"
    VISIBLE_PORTION = "visible portion"


@dataclass(kw_only=True)
class ScreenshotTempDataHandler():
    """Class to handle iteration details and general data for screenshot of complete page, e.g. file and directory names."""

    __slots__ = ["destination_dir", "destination_file_path",
                 "_all_temp_file_paths", "_temp_file_prefix", "_iteration", "_temp_dir"]

    destination_dir: str
    destination_file_path: str

    def __post_init__(self) -> None:
        self.destination_file_path = FilePNG(self.destination_file_path)
        self._temp_dir: str = helper.screenshot.controller.mediate_temp_dir(self.destination_dir)
        self._all_temp_file_paths: list[str] = []
        self._temp_file_prefix: str = helper.screenshot.file.get_temp_prefix_without_iterator_and_file_type()
        self._iteration: int = 1

    def get_temp_file_name(self) -> str:
        return f"{self._temp_file_prefix}_{self._iteration}.png"

    def get_temp_file_path(self) -> str:
        temp_file_name = self.get_temp_file_name()
        return helper.screenshot.file.get_path(self._temp_dir, temp_file_name)

    async def save_screenshot(self, driver: object) -> None:
        temp_file_path = self.get_temp_file_path()
        helper.screenshot.save(driver, temp_file_path)
        self._all_temp_file_paths.append(temp_file_path)

    def increment_iteration(self) -> None:
        self._iteration += 1

    async def merge_temp_files_into_final_screenshot(self) -> None:
        helper.screenshot.merge_images(self._all_temp_file_paths, self.destination_file_path)

    def remove_temp_files(self) -> None:
        helper.file.remove(self._all_temp_file_paths)



from ....constant import timeout
from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from .value import get_attribute_value
from .values import get_attribute_values


class GetAttributeDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def value(self, xpath: str, attribute: str, timeout: float = timeout.DEFAULT) -> str:
        """Get value from an attribute of an element. Examples:

        Use "src" as attribute to get the source URL from an <img> image tag.

        Use "href" as attribute to get the URL from an <a> link tag."""

        return get_attribute_value(self._driver, xpath, attribute, timeout)

    def values(self, xpath: str, attribute: str, timeout: float = timeout.DEFAULT) -> list[str]:
        """Get values from an attribute of multiple elements. Assumes that the XPath targets multiple links. Examples:

        Use "src" as attribute to get the source URL from an <img> image tag.

        Use "href" as attribute to get the URL from an <a> link tag."""

        return get_attribute_values(self._driver, xpath, attribute, timeout)



from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait  # type: ignore

from .....constant import timeout
from .....exception.timeout import WaitForUrlTimeoutException


def wait_until_url_contains(driver: object, url_fragment: str, timeout: float = timeout.DEFAULT) -> None:
    try:
        WebDriverWait(driver, timeout).until(EC.url_contains(url_fragment))  # type: ignore
    except TimeoutException:
        raise WaitForUrlTimeoutException(driver, url_fragment) from TimeoutException



from ..browser.get.url.current import get_current_url


class NoElementFoundException(Exception):
    __slots__ = ["message"]

    def __init__(self, driver: object, xpath: str) -> None:
        current_url = get_current_url(driver)
        self.message = f"On page {current_url}, no such element found while waiting: {xpath}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class NoElementFoundWithTextConditionException(Exception):
    __slots__ = ["message"]

    def __init__(self, driver: object, xpath: str, text_condition: str) -> None:
        current_url = get_current_url(driver)
        self.message = f"On page {current_url}, an element that contains \"{text_condition}\" could not be found: {xpath}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class NoElementDimensionsFoundException(Exception):
    __slots__ = ["message"]

    def __init__(self, driver: object, xpath: str) -> None:
        current_url = get_current_url(driver)
        self.message = f"On page {current_url}, no such element found while trying to get element dimensions: {xpath}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



__all__ = ["Browser", "BrowserSettings", "BrowserType", "CookieBannerSettings", "LoginCredentials",
           "LoginForm1Step", "LoginForm2Steps", "PageLoadStrategy", "SearchSettings", "TimeoutSettings", "TimeoutStrategy"]

from .browser.__main__ import Browser
from .model.browser.base.page_load_strategy import PageLoadStrategy
from .model.browser.base.settings import BrowserSettings
from .model.browser.base.type import BrowserType
from .model.combo_settings.cookie_banner import CookieBannerSettings
from .model.combo_settings.login_credentials import LoginCredentials
from .model.combo_settings.login_form import LoginForm1Step, LoginForm2Steps
from .model.combo_settings.search import SearchSettings
from .model.timeout_strategy import TimeoutSettings, TimeoutStrategy
from .version import __version__  # noqa



from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from .position import get_scroll_position
from .total_height import get_total_scroll_height


class ScrollGetDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def position(self) -> tuple[int, int]:
        """Get scroll position of the X and Y axis. Usage:

        x, y = browser.scroll.get.position()"""

        return get_scroll_position(self._driver)

    def total_height(self) -> int:
        """Get total scroll height."""

        return get_total_scroll_height(self._driver)



from .....constant import timeout
from .....model.browser.base.driver import BrowserDriver
from .....model.browser.base.settings import BrowserSettings
from .....model.driver_methods import DriverMethods
from .changes import wait_until_text_changes
from .contains import wait_until_text_contains
from .equals import wait_until_text_equals


class WaitUntilTextDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def changes(self, xpath: str, baseline_text: str, timeout: float = timeout.DEFAULT) -> None:
        """Wait until the text of an element changes from a baseline text, e.g. after a form action. The text is evaluated as an exact match."""

        wait_until_text_changes(self._driver, xpath, baseline_text, timeout)

    def contains(self, xpath: str, regex: str, timeout: float = timeout.DEFAULT) -> None:
        """Wait until the text of an element has changed, e.g. after a form action."""

        wait_until_text_contains(self._driver, xpath, regex, timeout)

    def equals(self, xpath: str, regex: str, timeout: float = timeout.DEFAULT) -> None:
        """Wait until the text of an element has changed, e.g. after a form action. The text is evaluated as an exact match."""

        wait_until_text_equals(self._driver, xpath, regex, timeout)



from selenium.webdriver.common.by import By

from .... import helper
from ....constant import timeout
from ....model.type.xpath import XPath
from ..for_element import wait_for_element


def wait_until_images_have_loaded(driver: object, xpath: str, timeout: float = timeout.DEFAULT) -> None:
    def are_all_images_loaded(driver: object, elements: list[object]) -> bool:
        return all(helper.image.is_element_loaded(driver, element) is not False for element in elements)

    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    elements: list[object] = driver.find_elements(By.XPATH, xpath)  # type: ignore
    helper.retry.until_condition_is_true(driver, elements, func=are_all_images_loaded, timeout=timeout)



from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class WindowHandle:
    """Object that let's the user define a humanised name for a tab/window while the actual ID is handled by the system in the background."""

    name: str
    id: str



from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait  # type: ignore

from .....constant import timeout
from .....exception.timeout import WaitForUrlTimeoutException
from .....model.type.url import URL


def wait_until_url_equals(driver: object, url: str, timeout: float = timeout.DEFAULT) -> None:
    url = URL(url)
    try:
        WebDriverWait(driver, timeout).until(EC.url_matches(url))  # type: ignore
    except TimeoutException:
        raise WaitForUrlTimeoutException(driver, url) from TimeoutException



from ...constant import timeout
from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .for_element import wait_for_element
from .random_time import wait_random_time
from .until.__main__ import WaitUntilDriverMethods


class WaitDriverMethods(DriverMethods):
    __slots__ = ["until"]

    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)
        self.until: WaitUntilDriverMethods = WaitUntilDriverMethods(browser_driver, settings)

    def for_element(self, xpath: str, timeout: float = timeout.DEFAULT) -> None:
        """Wait until element is ready in the DOM and/or on the screen.

        Especially useful for single-page app elements handled/modified by JavaScript, but also standard HTML that doesn't load immediately, this helper function ensures that DOM elements are ready before processing."""

        wait_for_element(self._driver, xpath, timeout)

    def random_time(self, min_seconds: float = 1, max_seconds: float = 5) -> None:
        """Randomize sleep timing to make actions look less like a bot."""

        wait_random_time(min_seconds, max_seconds)
